import logging
import warnings
from contextlib import contextmanager

from rosnik import config, state

logger = logging.getLogger(__name__)

openai_enabled = False

try:
    from rosnik.providers import openai as openai_

    openai_enabled = True
except ImportError:
    pass


def init(api_key=None, sync_mode=None, environment=None, event_context_hook=None):
    config.Config.api_key = api_key
    config.Config.sync_mode = sync_mode
    config.Config.environment = environment
    config.Config.event_context_hook = event_context_hook

    if config.Config.api_key is None:
        warnings.warn("`api_key` is not set and an API token was not provided on init")

    if openai_enabled:
        logger.debug("OpenAI is enabled. Patching.")
        openai_._patch_openai()
    else:
        logger.debug("Skipping OpenAI instrumentation.")

    if config.Config.sync_mode:
        logger.debug("Running in sync mode")


@contextmanager
def context(prompt_name: str = None, **kwargs):
    """Set context for the events sent within this context manager."""
    token = state.store(state.State.CONTEXT_ID, {"prompt_name": prompt_name, **kwargs})
    yield
    state.reset_context(token)



from .client import init, context
from .frameworks import flask_rosnik, django
from .events.user import track_user_feedback, track_user_interaction

__all__ = [
    "init",
    "flask_rosnik",
    "track_user_interaction",
    "track_user_feedback",
    "django",
    "context",
]



import logging
import threading
import queue
import os

from rosnik import api
from rosnik import config
from rosnik.types.core import Event

logger = logging.getLogger(__name__)


class EventProcessor(threading.Thread):
    def __init__(self, queue: queue.Queue, api_client: api.IngestClient):
        super().__init__()
        self.daemon = True
        self.started = False
        self.event_queue = queue
        self.api_client = api_client
        self.pid = os.getpid()

    def run(self):
        self.started = True
        self.process_events()

    def process_events(self):
        while True:
            event = self.event_queue.get()
            self.api_client.send_event(event)


event_queue = queue.Queue()
api_client = api.IngestClient()
event_processor = None


def enqueue_event(event: Event):
    if config.Config.sync_mode:
        logger.debug(f"Enqueuing event in sync mode: {event.event_id}")
        api_client.send_event(event)
        return

    global event_processor
    # If we're sending events in the background, start this thread.
    if event_processor is None:
        logger.debug("Event processor is None. Starting a new one.")
        event_processor = EventProcessor(event_queue, api_client)
        event_processor.start()

    try:
        logger.debug(f"Enqueuing event {event.event_id}")
        event_queue.put(event, block=False)
    except queue.Full:
        logger.warning("rosnik events queue is full")



import logging
import warnings

from rosnik import client, headers, state

try:
    from flask import request
except ImportError:
    request = None

logger = logging.getLogger(__name__)


def _before_request():
    # Import this JIT to avoid warning / errors for non-Flask environments.
    if request is None:
        warnings.warn("Could not import flask.request")

    journey_id = request.headers.get(headers.JOURNEY_ID_KEY)
    if journey_id is None:
        journey_id = state.create_journey_id()
    state.store(state.State.JOURNEY_ID, journey_id)

    interaction_id = request.headers.get(headers.INTERACTION_ID_KEY)
    state.store(state.State.USER_INTERACTION_ID, interaction_id)

    device_id = request.headers.get(headers.DEVICE_ID_KEY)
    state.store(state.State.DEVICE_ID, device_id)


def _annotate_response_headers(response):
    response.headers[headers.JOURNEY_ID_KEY] = state.get_journey_id()
    return response


class FlaskRosnik:
    def __init__(
        self, app=None, api_key=None, sync_mode=None, environment=None, event_context_hook=None
    ):
        self.api_key = api_key
        self.sync_mode = sync_mode
        self.environment = environment
        self.event_context_hook = event_context_hook
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        app.before_request(_before_request)
        app.after_request(_annotate_response_headers)
        client.init(
            api_key=self.api_key,
            sync_mode=self.sync_mode,
            environment=self.environment,
            event_context_hook=self.event_context_hook,
        )



import inspect
import logging
import sys
from typing import Callable


import wrapt

logger = logging.getLogger(__name__)


def get_stack_frames(num, use_get_frame=True):
    """Quickly get stack frames via:
    https://gist.github.com/csm10495/39dde7add5f1b1e73c4e8299f5df1116
    """
    # Not all versions of python have the sys._getframe() method.
    # All should have inspect, though it is really slow
    if use_get_frame and hasattr(sys, "_getframe"):
        frame = sys._getframe(0)
        frames = [frame]

        while frame.f_back is not None and len(frames) < num:
            frames.append(frame.f_back)
            frame = frame.f_back

        return frames
    else:
        return inspect.stack()[:num]


def wrap_class_method(
    klass,
    method_name: str,
    request_hook: Callable,
    response_hook: Callable,
    error_hook: Callable,
    streamed_response_hook: Callable,
):
    def rosnik_wrapper(wrapped, instance, args, kwargs):
        limited_frames = get_stack_frames(10)
        # Flatten into a period separated sequence so we can do function chain search later.
        calling_functions = ".".join([frame.f_code.co_name for frame in limited_frames])

        request_event = request_hook(kwargs, calling_functions)
        try:
            result = wrapped(*args, **kwargs)
        except Exception as e:
            error_hook(e, calling_functions, request_event)
            raise e

        response_hook(result, calling_functions, prior_event=request_event)

        # If this is a streamed output, wrap this so we can capture stream duration
        # and final output.
        if kwargs.get("stream") is True:
            return streamed_response_hook(result, calling_functions, prior_event=request_event)

        return result

    wrapt.wrap_function_wrapper(klass, method_name, rosnik_wrapper)



from rosnik import client, headers, state

try:
    from django.conf import settings
except ImportError:
    settings = None


def _to_django_header(key):
    """Converts our header key to a Django header key."""
    return f'HTTP_{key.replace("-", "_").upper()}'


def rosnik_middleware(get_response):
    # Initialize our SDK at the start using Django settings.
    client.init(
        api_key=getattr(settings, "ROSNIK_API_KEY", None),
        sync_mode=getattr(settings, "ROSNIK_SYNC_MODE", None),
        environment=getattr(settings, "ROSNIK_ENVIRONMENT", None),
        event_context_hook=getattr(settings, "ROSNIK_EVENT_CONTEXT_HOOK", None),
    )

    def middleware(request):
        """Each request is considered a distinct Journey, unless a Journey ID is supplied."""
        journey_id = request.META.get(_to_django_header(headers.JOURNEY_ID_KEY))
        if journey_id is None:
            journey_id = state.create_journey_id()
        state.store(state.State.JOURNEY_ID, journey_id)

        interaction_id = request.META.get(_to_django_header(headers.INTERACTION_ID_KEY))
        state.store(state.State.USER_INTERACTION_ID, interaction_id)

        device_id = request.META.get(_to_django_header(headers.DEVICE_ID_KEY))
        state.store(state.State.DEVICE_ID, device_id)

        response = get_response(request)

        # Pass our Journey ID back to the client,
        # so we can use it for subsequent requests.
        response[headers.JOURNEY_ID_KEY] = state.get_journey_id()
        return response

    return middleware



"""
A request on its own can be considered a journey,
if there is not an external journey ID supplied.

We manage state here across the 3 values we care about.
Outside of Journey ID, they can be set to None, which
would mean they are user-less journeys.

We also manage context here, which is a user-defined
JSONable dictionary of key-value pairs that get
collected and sent with every event. See `rosnik.context`
for the context manager and details on how to use it.
"""
import contextvars
from enum import Enum
import logging
from contextvars import ContextVar
from typing import Optional

import ulid

logger = logging.getLogger(__name__)

_journey_id_key = "journey_id"
_user_interaction_id_key = "user_interaction_id"
_device_id_key = "device_id"
_context_key = "context"

journey_id_cv: ContextVar[Optional[str]] = ContextVar(_journey_id_key)
user_interaction_id_cv: ContextVar[Optional[str]] = ContextVar(_user_interaction_id_key)
device_id_cv: ContextVar[Optional[str]] = ContextVar(_device_id_key)
context_cv: ContextVar[Optional[dict[str, str]]] = ContextVar(_context_key)


class State(Enum):
    JOURNEY_ID = _journey_id_key
    USER_INTERACTION_ID = _user_interaction_id_key
    DEVICE_ID = _device_id_key
    CONTEXT_ID = _context_key


_state = {
    State.JOURNEY_ID: journey_id_cv,
    State.USER_INTERACTION_ID: user_interaction_id_cv,
    State.DEVICE_ID: device_id_cv,
    State.CONTEXT_ID: context_cv,
}


def store(state_type: State, value: str):
    return _state[state_type].set(value)


def retrieve(state_type: State):
    return _state[state_type].get(None)


def create_journey_id():
    journey_id = ulid.new().str
    store(State.JOURNEY_ID, journey_id)
    return journey_id


def get_journey_id():
    journey_id = retrieve(State.JOURNEY_ID)
    if journey_id is None:
        return create_journey_id()
    return journey_id


def get_device_id():
    return retrieve(State.DEVICE_ID)


def get_user_interaction_id():
    return retrieve(State.USER_INTERACTION_ID)


def get_context():
    return retrieve(State.CONTEXT_ID) or {}


def reset_context(token: contextvars.Token):
    return context_cv.reset(token)


def _reset():
    journey_id_cv.set(None)
    user_interaction_id_cv.set(None)
    device_id_cv.set(None)
    context_cv.set(None)



from dataclasses import dataclass
from typing import Optional

from dataclasses_json import DataClassJsonMixin


from rosnik.types.core import AIEvent


@dataclass(kw_only=True, slots=True)
class OpenAIAttributes(DataClassJsonMixin):
    api_base: str
    api_type: str
    api_version: str
    organization: Optional[str] = None
    # OpenAI tracks response ms
    response_ms: Optional[int] = None


@dataclass(kw_only=True, slots=True)
class ResponseData(DataClassJsonMixin):
    response_payload: dict
    organization: str
    response_ms: int
    openai_attributes: OpenAIAttributes


@dataclass(kw_only=True, slots=True)
class ErrorResponseData(DataClassJsonMixin):
    message: str
    json_body: Optional[str] = None
    headers: Optional[dict] = None
    http_status: Optional[int] = None
    organization: Optional[str] = None
    request_id: Optional[str] = None
    http_body: Optional[str] = None
    user_message: Optional[str] = None
    code: Optional[str] = None


@dataclass(kw_only=True, slots=True)
class AIFunctionMetadata(DataClassJsonMixin):
    ai_provider: str
    ai_action: str
    openai_attributes: Optional[OpenAIAttributes] = None


@dataclass(kw_only=True, slots=True)
class AIRequestStart(AIEvent):
    event_type: str = "ai.request.start"
    # JSONable request payload / kwargs sent through the SDK
    request_payload: dict
    ai_metadata: AIFunctionMetadata
    ai_request_start_event_id: Optional[str] = None

    def __post_init__(self):
        AIEvent.__post_init__(self)
        self.ai_request_start_event_id = self.event_id


@dataclass(kw_only=True, slots=True)
class AIRequestFinish(AIEvent):
    event_type: str = "ai.request.finish"
    # JSONable response payload/body
    # null on errors
    response_payload: Optional[dict] = None
    # This is the Event ID of the ai.request.start
    # event that caused this finish action.
    ai_request_start_event_id: str
    ai_metadata: AIFunctionMetadata
    # Null on error
    # Our client-side calculation of duration
    response_ms: int
    # null on success
    error_data: Optional[ErrorResponseData] = None


@dataclass(kw_only=True, slots=True)
class AIRequestStartStream(AIRequestFinish):
    """One-off event used to track first-byte time."""

    event_type: str = "ai.request.start.stream"


@dataclass(kw_only=True, slots=True)
class AIOperationStart(AIEvent):
    pass


@dataclass(kw_only=True, slots=True)
class AIOperationFinish(AIEvent):
    pass



from typing import Optional
from rosnik.types.user import UserFeedbackTrack, UserInteractionTrack
from rosnik.events import queue


def track_user_interaction(user_id: str, interaction_type: str):
    event = UserInteractionTrack(user_id=user_id, interaction_type=interaction_type)
    queue.enqueue_event(event)
    return event


def track_user_feedback(
    user_id: str, score: Optional[int] = None, open_response: Optional[str] = None
):
    event = UserFeedbackTrack(user_id=user_id, score=score, open_response=open_response)
    queue.enqueue_event(event)
    return event



from rosnik import constants

JOURNEY_ID_KEY = f"X-{constants.NAMESPACE}-Journey-Id"
INTERACTION_ID_KEY = f"X-{constants.NAMESPACE}-Interaction-Id"
DEVICE_ID_KEY = f"X-{constants.NAMESPACE}-Device-Id"



import logging
import time
import platform
from dataclasses import dataclass, field
from typing import Callable, Optional

from dataclasses_json import DataClassJsonMixin
import ulid

from rosnik import config
from rosnik import state

logger = logging.getLogger(__name__)


def _generate_event_id():
    return ulid.new().str


@dataclass(kw_only=True, slots=True)
class StaticMetadata:
    environment: Optional[str] = field(default_factory=lambda: config.Config.environment)
    runtime: str = platform.python_implementation()
    runtime_version: str = platform.python_version()
    # TODO: how to sync pyproject version to this
    sdk_version: str = "0.0.36"


@dataclass(kw_only=True, slots=True)
class Metadata(StaticMetadata):
    function_fingerprint: str
    stream: bool = False


_reserved_words = ["environment"]


@dataclass(kw_only=True, slots=True)
class Event(DataClassJsonMixin):
    # Unique event ID
    event_id: str = field(default_factory=_generate_event_id)
    event_type: str
    journey_id: str = field(default_factory=state.get_journey_id)
    # Epoch in ms
    sent_at: int = field(default_factory=lambda: int(time.time_ns() / 1000000))
    # Epoch unless not set, which will be -1
    occurred_at: Optional[int] = None
    # JSONable user defined context
    context: Optional[dict] = None
    # Users could be part of an AI event or a User event
    user_id: Optional[str] = None
    device_id: Optional[str] = field(default_factory=state.get_device_id)
    # Our own metadata
    _metadata: Metadata
    # User Interaction ID: this is the causal user.interaction.track
    # event ID for this event. If it's unset then something else
    # we're not tracking caused this action.
    user_interaction_id: Optional[str] = field(default_factory=state.get_user_interaction_id)

    def __post_init__(self):
        """We have two different contexts:
        1. Hook-level context which request-level / global
        2. Event-level context which is per-event / via a context manager
        Here we merge them together.
        """
        stored_context = state.get_context()

        # If both contexts are empty, we don't need to do anything.
        if not isinstance(config.Config.event_context_hook, Callable) and not stored_context:
            return

        # Make this a callable so we can call it later if needed.
        event_hook = config.Config.event_context_hook
        if not isinstance(event_hook, Callable):

            def event_hook():
                return {}

        try:
            global_context = event_hook()
            # Following the same precedence as below, pull environment if set.
            # Event-level context takes precedence over stored context, which takes precedence
            # over global context. If none of them have an environment, then we don't set it.
            env = global_context.pop("environment", None)
            if env:
                self._metadata.environment = env

            env = stored_context.pop("environment", None)
            if env:
                self._metadata.environment = env

            if self.context is not None:
                env = self.context.pop("environment", None)
                if env:
                    self._metadata.environment = env

            # Merge everything together with this event's specific context taking precedence,
            # then the stored context, then the global context.
            self.context = {**global_context, **stored_context, **(self.context or {})}
        except Exception:
            logger.exception(
                f"Could not generate context from {config.Config.event_context_hook.__name__}"
            )


@dataclass(kw_only=True, slots=True)
class AIEvent(Event):
    # model name
    ai_model: str
    # openai
    ai_provider: str
    # AI action: completion, chatcompletion
    ai_action: str


class UserEvent(Event):
    # JSONable user properties that can be added
    user_properties: dict
